#include <boost/config.hpp>
#include <iostream>
#include <algorithm>
#include <string.h>
#include <vector>

#include <boost/tuple/tuple.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/depth_first_search.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/pending/indirect_cmp.hpp>
#include <boost/pending/integer_range.hpp>
#include <boost/graph/connected_components.hpp>


#include <cstdlib>
#include <cstring>

using boost::adjacency_list;
using namespace std;
using namespace boost;

#include "graph_op.hpp"

extern "C" int surrogate(PACKED_GRAPH *);
extern "C" char **_breadth_first_search(PACKED_GRAPH *pg, char *name);
extern "C" char **_depth_first_search(PACKED_GRAPH *pg);
extern "C" char*** _connected_components(PACKED_GRAPH *pg, int *nc);

typedef boost::property<vertex_name_t, std::string> NameProperty;
typedef boost::property<edge_name_t,   std::string> EdgeProperty;
typedef boost::adjacency_list<vecS, vecS, directedS, NameProperty, EdgeProperty> Graph;
typedef boost::adjacency_list<vecS, vecS, undirectedS, NameProperty, 
               EdgeProperty> UndirectedGraph; 

typedef graph_traits<Graph>::vertex_descriptor VertexDescriptor;
typedef graph_traits<Graph>::edge_descriptor   EdgeDescriptor;
typedef graph_traits<Graph>::vertex_iterator   VertexIterator;
typedef graph_traits<Graph>::edge_iterator     EdgeIterator;

typedef graph_traits<Graph>::vertices_size_type Size;
typedef Size* Iiter;

template <typename TimeMap>
class bfs_time_visitor : public default_bfs_visitor {
   typedef typename property_traits<TimeMap>::value_type T;
public:
   bfs_time_visitor(TimeMap tmap, T& t) : m_timemap(tmap), m_time(t) {}
   template <typename Vertex, typename Graph>
   void discover_vertex(Vertex u, const Graph& g) const { 
      put(m_timemap, u, m_time++);
   }
   TimeMap m_timemap;
   T& m_time;
};

template < typename TimeMap > class dfs_time_visitor:public default_dfs_visitor {
  typedef typename property_traits < TimeMap >::value_type T;
public:
  dfs_time_visitor(TimeMap dmap, TimeMap fmap, T & t)
:  m_dtimemap(dmap), m_ftimemap(fmap), m_time(t) {
  }
  template < typename Vertex, typename Graph >
    void discover_vertex(Vertex u, const Graph & g) const
  {
    put(m_dtimemap, u, m_time++);
  }
  template < typename Vertex, typename Graph >
    void finish_vertex(Vertex u, const Graph & g) const
  {
    put(m_ftimemap, u, m_time++);
  }
  TimeMap m_dtimemap;
  TimeMap m_ftimemap;
  T & m_time;
};



tuple<VertexDescriptor,int> FindVertex( std::string name, Graph g )
{
   property_map<Graph, vertex_name_t>::type vnames = get(vertex_name, g);
   VertexIterator iter, iter_end;
   int v=0;

   tie(iter,iter_end) = vertices(g);
   for(;iter!=iter_end;iter++)
      if ( get(vnames,*iter)==name ) { v=1; break; }

   return tie(*iter,v);
   
}

/*
tuple<EdgeDescriptor,int> FindEdge( VertexDescriptor e1, VertexDescriptor e2, Graph g)
{
   property_map<Graph, edge_name_t>::type enames = get( edge_name, g);
   EdgeIterator iter, iter_end;
   int v;

   v = 0;
   tie(iter,iter_end) = edges(g);
   for(;iter!=iter_end;iter++)
      if ( (source(*iter,g)==e1) && (target(*iter,g)==e2) )  { v=1; break; }

   return tie(*iter,v);
}
*/

UndirectedGraph utransform(PACKED_GRAPH *graph)
{
   int num;
   int length, len, source, destination, i, n1,n2;
   unsigned int *offset;

   num = PACKED_GRAPH_NUM_NODES(graph);

   UndirectedGraph g(num);
   property_map<UndirectedGraph, vertex_name_t>::type vnames = get(vertex_name, g);
   property_map<UndirectedGraph, edge_name_t>::type enames = get( edge_name, g);

   vector<string> names(num);

   char *src = PACKED_GRAPH_NODES(graph);
   length = 0;

   // process vertex names
   for(i=0;i<num;i++)
   {
      names[i] = src;
      len = ::strlen(src)+1;
      length += len;
      src += len;
   }
   num = PACKED_GRAPH_NUM_EDGES(graph);
   offset = (unsigned int *) PACKED_GRAPH_EDGES(graph);
   for(i=0;i<num;i++)
   {
       source = *offset;
       destination = *(offset+1);
       add_edge( source, destination, g );
       offset +=2;
   }

   num = PACKED_GRAPH_NUM_NODES(graph);
   for(i=0;i<num;i++)
      vnames[i] = names[i];

   typedef graph_traits<UndirectedGraph>::vertex_iterator iter;
   typedef graph_traits<UndirectedGraph>::edge_iterator eiter;
   std::pair<iter,iter> vp = vertices(g);
   std::pair<eiter,eiter> ep;
   eiter p;
   typedef graph_traits<UndirectedGraph>::vertex_descriptor VD;
   typedef std::pair<VD,VD> E;

   ep = edges(g);

   num = PACKED_GRAPH_NUM_LABELS(graph);
   src = PACKED_GRAPH_LABELS(graph);
   for(i=0;i<num;i++)
   {
      offset = (unsigned int *) src;
      n1 = *offset;
      n2 = *(offset+1);
      src += (2*sizeof(int));

      for(p=ep.first;p!=ep.second;p++)
         if ( E(boost::source(*p,g),boost::target(*p,g)) == 
              E(vp.first[n1],vp.first[n2]))  break;


      if ( p!=ep.second )
      {
         std::string label(src);
         enames[*p] = label;
         cout << "edge:" << *p << " has label: " << label << endl;
      }
      else
         cout << "could not happen" << endl;

      len = ::strlen(src);
     src += (len+1);
   }

   typedef graph_traits<UndirectedGraph>::vertex_iterator iter;
   std::pair<iter,iter> bp;
   for(bp=vertices(g); bp.first != bp.second; ++bp.first )
      std::cout << get(vnames,*bp.first) << endl;

   return g;
}


Graph transform(PACKED_GRAPH *graph)
{
   int num;
   int length, len, source, destination, i, n1,n2;
   unsigned int *offset;
   
   num = PACKED_GRAPH_NUM_NODES(graph);

   Graph g(num);
   property_map<Graph, vertex_name_t>::type vnames = get(vertex_name, g);  
   property_map<Graph, edge_name_t>::type enames = get( edge_name, g);
   
   vector<string> names(num);

   char *src = PACKED_GRAPH_NODES(graph);
   length = 0;

   // process vertex names
   for(i=0;i<num;i++)
   {
      names[i] = src;
      len = ::strlen(src)+1;
      length += len;
      src += len;      
   }

   num = PACKED_GRAPH_NUM_EDGES(graph);
   offset = (unsigned int *) PACKED_GRAPH_EDGES(graph);
   for(i=0;i<num;i++)
   {
       source = *offset;
       destination = *(offset+1);
       add_edge( source, destination, g );
       offset +=2;
   }
   
   num = PACKED_GRAPH_NUM_NODES(graph);
   for(i=0;i<num;i++)
      vnames[i] = names[i];

   typedef graph_traits<Graph>::vertex_iterator iter;
   typedef graph_traits<Graph>::edge_iterator eiter;  
   std::pair<iter,iter> vp = vertices(g);
   std::pair<eiter,eiter> ep;
   eiter p;
   typedef graph_traits<Graph>::vertex_descriptor VD;
   typedef std::pair<VD,VD> E;

   ep = edges(g);

   num = PACKED_GRAPH_NUM_LABELS(graph);
   src = PACKED_GRAPH_LABELS(graph);
   for(i=0;i<num;i++)
   {
      offset = (unsigned int *) src;
      n1 = *offset;
      n2 = *(offset+1);
      src += (2*sizeof(int));

      for(p=ep.first;p!=ep.second;p++)
         if ( E(boost::source(*p,g),boost::target(*p,g)) == E(vp.first[n1],vp.first[n2]) )  break;

      if ( p!=ep.second )
      {
         std::string label(src);
         enames[*p] = label;
         //cout << "edge:" << *p << " has label: " << label << endl;
      }
      //else
      //   cout << "could not happen" << endl;

      len = ::strlen(src);
      src += (len+1);
   }

   typedef graph_traits<Graph>::vertex_iterator iter;
   //std::pair<iter,iter> bp;
   //for(bp=vertices(g); bp.first != bp.second; ++bp.first )
   //   std::cout << get(vnames,*bp.first) << endl;   

   /*
   // out-edges
   typedef graph_traits<Graph>::out_edge_iterator OEI;
   OEI out, out_end;
   typedef graph_traits<Graph>::in_edge_iterator IEI;
   IEI oin, oin_end;

   std::cout << "execute out_edges..."  << endl; 
   VertexDescriptor aa;
   int bb;
   tie(aa,bb) = FindVertex("B",g);

   std::cout << "finding outedges.." << endl;
   if ( bb )
   {
   for(tie(out,out_end)=out_edges( aa,g); out!=out_end; ++out)
	std::cout << *out;
   }
   std::cout << endl;

   // start breadth-first search
   tie(aa,bb) = FindVertex("A",g);
   std::cout << "starting breadth-first-search with a" << endl;

   if (bb==1)
   {

   std::vector<Size> dtime(num_vertices(g));
   Size time = 0;
   bfs_time_visitor<Size*> vis(&dtime[0], time);
   breadth_first_search(g, aa, visitor(vis));

   // process the order
   int N = num_vertices(g);
   std::vector<graph_traits<Graph>::vertices_size_type> discover_order(N);
   integer_range<int> range(0,N);
   std::copy(range.begin(), range.end(), discover_order.begin());
   std::sort(discover_order.begin(), discover_order.end(), 
         indirect_cmp<Iiter, std::less<Size> >(&dtime[0]));
   std::cout << "discovery order:";
   for(i=0;i<N;i++)
     std::cout << vnames[ discover_order[i] ] << " ";
   }

   std::cout << "end..." << endl;


   // start depth-first search
   std::cout << "starting depth-first-search with a" << endl;


   std::vector<Size> dtime1(num_vertices(g));
   std::vector<Size> ftime(num_vertices(g));
   Size t = 0;
   dfs_time_visitor<Size*> vis(&dtime1[0], &ftime[0], t);
   depth_first_search(g, visitor(vis));

   // process the order
   int N = num_vertices(g);
   std::vector<graph_traits<Graph>::vertices_size_type> discover_order1(N);
   integer_range<int> range1(0,N);
   std::copy(range1.begin(), range1.end(), discover_order1.begin());
   std::sort(discover_order1.begin(), discover_order1.end(),
         indirect_cmp<Iiter, std::less<Size> >(&dtime1[0]));
   std::cout << "discovery order:";
   for(i=0;i<N;i++)
     std::cout << vnames[ discover_order1[i] ] << " ";

   std::cout << "end..." << endl;
   */


   return g;
}


char **_breadth_first_search(PACKED_GRAPH *pg, char *name )
{
   VertexDescriptor vd;
   int b;
   char **ret;

   ret = NULL;
   Graph g = transform(pg);
   property_map<Graph, vertex_name_t>::type vnames = get(vertex_name, g);

   tie(vd,b) = FindVertex( std::string(name),g );

   if (b==1)
   {
      std::vector<Size> dtime(num_vertices(g));
      Size time = 0;
      bfs_time_visitor<Size*> vis(&dtime[0], time);
      breadth_first_search(g, vd, visitor(vis));

      // process the order
      int N = num_vertices(g);
      std::vector<graph_traits<Graph>::vertices_size_type> discover_order(N);
      integer_range<int> range(0,N);
      std::copy(range.begin(), range.end(), discover_order.begin());
      std::sort(discover_order.begin(), discover_order.end(),
            indirect_cmp<Iiter, std::less<Size> >(&dtime[0]));

      N = dtime.size();

      ret = (char **) ::malloc( sizeof(char*) * (N+1) );
      for(int i=0;i<N;i++)
      {
         //std::cout << vnames[ discover_order[i] ] << " ";
         ret[i] = (char *) ::strdup( vnames[discover_order[i]].c_str() );
      } 
      ret[N] = (char *) NULL;
   }

   return ret;
}

//char **_depth_first_search(PACKED_GRAPH *pg, char *name)
char **_depth_first_search(PACKED_GRAPH *pg)
{
   //VertexDescriptor vd;
   //int b;
   char **ret;

   ret = NULL;
   Graph g = transform(pg);
   property_map<Graph, vertex_name_t>::type vnames = get(vertex_name, g);

   //tie(vd,b) = FindVertex( std::string(name),g );

   std::vector<Size> dtime(num_vertices(g));
   std::vector<Size> ftime(num_vertices(g));
   Size t = 0;
   dfs_time_visitor<Size*> vis(&dtime[0], &ftime[0], t);
   depth_first_search(g, visitor(vis));

   // process the order
   int N = num_vertices(g);
   std::vector<graph_traits<Graph>::vertices_size_type> discover_order(N);
   integer_range<int> range(0,N);
   std::copy(range.begin(), range.end(), discover_order.begin());
   std::sort(discover_order.begin(), discover_order.end(),
         indirect_cmp<Iiter, std::less<Size> >(&dtime[0]));

   //std::cout << "discovery order:";

   //for(i=0;i<N;i++)
   //  std::cout << vnames[ discover_order1[i] ] << " ";

   //std::cout << "end..." << endl;

   N = dtime.size();

   ret = (char **) ::malloc( sizeof(char*) * (N+1) );
   for(int i=0;i<N;i++)
   {
      ret[i] = (char *) ::strdup( vnames[discover_order[i]].c_str() );
   }
   ret[N] = (char *) NULL;

   return ret;
}


char*** _connected_components(PACKED_GRAPH *pg, int *nc)
{
   UndirectedGraph G = utransform(pg);
   property_map<UndirectedGraph,vertex_name_t>::type vnames = get(vertex_name, G);

   std::vector<int> c(num_vertices(G));
   int num = connected_components(G, 
         make_iterator_property_map(c.begin(), get(vertex_index,G)));

   *nc = num;
 
   std::cout << std::endl;
   std::vector<int>::iterator i;
   std::cout << "Total number of components: " << num << std::endl;
   
   typedef std::vector<int> my_vector;
   std::vector<my_vector> v( num );
   char ***ret = (char ***) malloc( sizeof(char **) * (num+1));

   for(i=c.begin(); i!=c.end(); ++i)
   {
      v[*i].push_back(i-c.begin());
   }


   vnames = get(vertex_name, G);
   for(int j=0;j<num;j++)
   {
      ret[j] = (char **) malloc( sizeof(char *) * (v[j].size()+1) );

      std::cout << "component " << j << " contains " <<
      v[j].size() << " elements :";

      for(i=v[j].begin(); i!=v[j].end(); ++i)
      {
          std::string a;
          a = (std::string) ( get(vnames, *i));
          std::cout << " " << a << " "; 
          ret[j][(int) (i-v[j].begin())] = (char *) strdup( a.c_str() );
      }
      ret[j][(int) (v[j].end()-v[j].begin())] = NULL;

      std::cout << std::endl;
   }

   ret[num] = NULL;

   return ret;
}



int surrogate(PACKED_GRAPH *g)
{
   Graph graph = transform(g);
   return 0;
}

