/*
**  PostGraph -- Graph data management in PostgreSQL
**
**  Copyright (C) 2005, 2006  Ruey-Lung Hsiao,  D.S. Parker
**  (rlhsiao@cs.ucla.edu, stott@cs.ucla.edu)
**
**  UCLA Center for Computational Biology
**  635 Charles Young Drive South, Suite 225
**  Los Angeles, CA 90095-7332  USA
**
**  UCLA Computer Science Dept.
**  3532 Boelter Hall
**  Los Angeles, CA 90095-1596  USA
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public License
**  as published by the Free Software Foundation; either version 2.1 of
**  the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**  See the GNU Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the
**  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
**  Boston, MA  02110-1301  USA
*/

#ifndef __GRAPH_H
#define __GRAPH_H

#ifdef __cplusplus
extern "C" {
#endif

#include "postgres.h"
#include "utils/builtins.h"

#include "graph_op.h"


typedef int32 OFFSET;
#define InvalidOffset  -1
#define IsOFFSETValid(o)   ((o)!=InvalidOffset)
#define IsOFFSETInvalid(o) ((o)==InvalidOffset)
#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))

#define OFFSET_OF_NODE_TABLE(graph)    ( (graph)->node_table_offset)
#define OFFSET_OF_LABEL_TABLE(graph)    ( (graph)->label_table_offset)
#define OFFSET_OF_NODES(graph)    ( (graph)->nodes_offset)
#define OFFSET_OF_EDGES(graph)    ( (graph)->edges_offset)
#define OFFSET_OF_LABELS(graph)    ( (graph)->labels_offset)

void GraphNodeSpecParser( UNPACKED_GRAPH *graph, char *str );
void GraphEdgeSpecParser( UNPACKED_GRAPH *graph, char *str );
void GraphLabelSpecParser( UNPACKED_GRAPH *graph, char *str );
UNPACKED_GRAPH *GraphSpecParser( char *spec );
Datum graph_in(PG_FUNCTION_ARGS);
Datum graph_out(PG_FUNCTION_ARGS);

char *add_string( char *string, int *capacity, int *length, char *data, int preserved );


#define PG_GETARG_GRAPH(x)  ((PACKED_GRAPH*) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(x))))

Datum graph_node_count( PG_FUNCTION_ARGS );
Datum graph_edge_count( PG_FUNCTION_ARGS );
Datum graph_label_count( PG_FUNCTION_ARGS );
Datum graph_indegree( PG_FUNCTION_ARGS );
Datum graph_outdegree( PG_FUNCTION_ARGS );
Datum graph_has_node( PG_FUNCTION_ARGS );
Datum graph_has_edge(PG_FUNCTION_ARGS);
Datum graph_has_edge_1(PG_FUNCTION_ARGS);
Datum graph_add_node(PG_FUNCTION_ARGS);
Datum graph_add_edge(PG_FUNCTION_ARGS);
Datum graph_add_label(PG_FUNCTION_ARGS);
Datum graph_add_graph(PG_FUNCTION_ARGS);

Datum graph_has_self_loop(PG_FUNCTION_ARGS);
Datum graph_breadth_first_search(PG_FUNCTION_ARGS);
Datum graph_depth_first_search(PG_FUNCTION_ARGS);
Datum graph_out_edges(PG_FUNCTION_ARGS);
Datum graph_in_edges(PG_FUNCTION_ARGS);
Datum graph_connected_components(PG_FUNCTION_ARGS);

#ifdef __cplusplus
}
#endif
#endif
